"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MigrationGenerator = void 0;
const fs_1 = __importDefault(require("fs"));
const assert = __importStar(require("assert"));
const utils_1 = require("./utils");
const NEW_LINE = '\n';
const EMPTY_STRING = '';
const SPACE_TAB = ' ';
const SINGLE_QUOTE = `'`;
const DOUBLE_QUOTE = `"`;
const PACKAGE_NAME = require('../package.json').name;
class TemplateBuilder {
    constructor(doubleQuote, tabsWidth) {
        this.template = [];
        this.doubleQuote = doubleQuote;
        this.tabsWidth = tabsWidth !== null && tabsWidth !== void 0 ? tabsWidth : 4;
        this.quote = doubleQuote ? DOUBLE_QUOTE : SINGLE_QUOTE;
    }
    clone() {
        return new TemplateBuilder(this.doubleQuote, this.tabsWidth);
    }
    writeLn(content = EMPTY_STRING) {
        this.template.push(...[content, NEW_LINE]);
        return this;
    }
    write(content) {
        this.template.push(content);
        return this;
    }
    spaceTabs(tabs) {
        this.template.push(new Array(tabs * this.tabsWidth).fill(SPACE_TAB).join(EMPTY_STRING));
        return this;
    }
    quoted(content) {
        return this.load(this.clone().write(this.quote).write(content).write(this.quote));
    }
    load(builder) {
        this.template.push(builder.build());
        return this;
    }
    reset() {
        this.template.length = 0;
        return this;
    }
    build() {
        return this.template.join(EMPTY_STRING);
    }
}
class TemplateFactory {
    constructor(pkgName, doubleQuote, tabsWidth) {
        this.pkgName = pkgName;
        this.builder = new TemplateBuilder(doubleQuote, tabsWidth);
    }
}
class TypescriptTemplateFactory extends TemplateFactory {
    constructor(pkgName, doubleQuote, tabsWidth) {
        super(pkgName, doubleQuote, tabsWidth);
    }
    createMethodBuilder(name) {
        return this.builder
            .clone()
            .spaceTabs(1)
            .writeLn(`async ${name}({ app, firestore }: IMigrationInput): Promise<void> {`)
            .spaceTabs(2)
            .writeLn(`return Promise.resolve(undefined);`)
            .spaceTabs(1)
            .writeLn('}');
    }
    createImportsBuilder(imports) {
        return this.builder
            .clone()
            .write(`import { ${imports.join(', ')} } from `)
            .quoted(`${this.pkgName}/lib/types`)
            .writeLn(';');
    }
    createReversibleMigration(timestamp) {
        const [up, down] = ['up', 'down'].map((method) => this.createMethodBuilder(method));
        return this.builder
            .reset()
            .load(this.createImportsBuilder(['IMigrationInput', 'IReversibleMigration']))
            .writeLn()
            .writeLn(`export default class ReversibleMigration${timestamp} implements IReversibleMigration {`)
            .load(up)
            .writeLn()
            .load(down)
            .writeLn('}')
            .build();
    }
    createIrreversibleMigration(timestamp) {
        return this.builder
            .reset()
            .load(this.createImportsBuilder(['IIrreversibleMigration', 'IMigrationInput']))
            .writeLn()
            .writeLn(`export default class IrreversibleMigration${timestamp} implements IIrreversibleMigration {`)
            .load(this.createMethodBuilder('execute'))
            .writeLn('}')
            .build();
    }
}
class JavascriptTemplateFactory extends TemplateFactory {
    constructor(pkgName, doubleQuote, tabsWidth) {
        super(pkgName, doubleQuote, tabsWidth);
    }
    createIrreversibleMigration(timestamp) {
        return this.builder.
            reset()
            .writeLn(`module.exports = class IrreversibleMigration${timestamp} {`)
            .spaceTabs(1)
            .writeLn(`async execute({ app, firestore }) {`)
            .spaceTabs(2)
            .writeLn(`return Promise.resolve(undefined);`)
            .spaceTabs(1)
            .writeLn('}')
            .writeLn('}')
            .build();
    }
    createReversibleMigration(timestamp) {
        return this.builder.
            reset()
            .writeLn(`module.exports = class ReversibleMigration${timestamp} {`)
            .spaceTabs(1)
            .writeLn(`async up({ app, firestore }) {`)
            .spaceTabs(2)
            .writeLn(`return Promise.resolve(undefined);`)
            .spaceTabs(1)
            .writeLn('}')
            .writeLn()
            .spaceTabs(1)
            .writeLn(`async down({ app, firestore }) {`)
            .spaceTabs(2)
            .writeLn(`return Promise.resolve(undefined);`)
            .spaceTabs(1)
            .writeLn('}')
            .writeLn('}')
            .build();
    }
}
class MigrationGenerator {
    constructor(options) {
        this.options = options;
        this.factories = new Map();
        this.getMigrationName = (timestamp, name, ext) => `${timestamp}-${name}.${ext}`;
        this.factories.set('js', new JavascriptTemplateFactory(PACKAGE_NAME, !!options.doubleQuote, options.tabs));
        this.factories.set('ts', new TypescriptTemplateFactory(PACKAGE_NAME, !!options.doubleQuote, options.tabs));
    }
    static create(options) {
        return new MigrationGenerator(options);
    }
    get templateFactory() {
        switch (this.options.ext) {
            case 'js':
                // TODO: create the factory implementation for js templates
                return this.factories.get('js');
            case 'ts':
                return this.factories.get('ts');
            default:
                return this.factories.get('ts');
        }
    }
    generate(name) {
        var _a;
        const timestamp = Date.now();
        const irreversible = !!this.options['irreversible'];
        if (this.options.clone) {
            const [cloneFileName] = utils_1.getMigrationsFiles({ migrationsDir: this.options.path, searchString: this.options.clone });
            assert.ok(cloneFileName, `Filed to find migration file timestamp: ${this.options.clone}`);
            const ext = utils_1.getExtName(cloneFileName);
            const migrationName = this.getMigrationName(timestamp, name, ext);
            const clonePath = utils_1.getAbsolutePath({ migrationsDir: this.options.path, migrationFileName: cloneFileName });
            const migrationPath = utils_1.getAbsolutePath({ migrationsDir: this.options.path, migrationFileName: migrationName });
            const cloneFile = fs_1.default
                .readFileSync(clonePath)
                .toString()
                .replace(/(class\s+?)(Reversible|Irreversible)(Migration)([0-9]+)/s, `$1$2$3${timestamp}`);
            fs_1.default.writeFileSync(migrationPath, Buffer.from(cloneFile));
            console.log(`New migration clone created: ${migrationPath}`);
            return;
        }
        const migrationName = this.getMigrationName(timestamp, name, (_a = this.options.ext) !== null && _a !== void 0 ? _a : 'ts');
        const migrationPath = utils_1.getAbsolutePath({ migrationsDir: this.options.path, migrationFileName: migrationName });
        const template = irreversible
            ? this.templateFactory.createIrreversibleMigration(timestamp)
            : this.templateFactory.createReversibleMigration(timestamp);
        fs_1.default.writeFileSync(migrationPath, Buffer.from(template));
        console.log(`New migration file created: ${migrationPath}`);
    }
}
exports.MigrationGenerator = MigrationGenerator;
